
import org.sosy_lab.common.ShutdownManager;
import org.sosy_lab.common.configuration.Configuration;
import org.sosy_lab.common.configuration.InvalidConfigurationException;
import org.sosy_lab.common.log.BasicLogManager;
import org.sosy_lab.common.log.LogManager;
import org.sosy_lab.java_smt.SolverContextFactory;
import org.sosy_lab.java_smt.SolverContextFactory.Solvers;
import org.sosy_lab.java_smt.api.*;
import org.sosy_lab.java_smt.api.NumeralFormula.IntegerFormula;
import org.sosy_lab.java_smt.api.SolverContext.ProverOptions;
import java.util.List;
import java.util.ArrayList;

/**
 * A simple riddle taken from
 * http://puzzleparasite.blogspot.de/2015/04/2014-slovak-puzzle-championships.html
 */
public class Riddle {
  public static void main(String[] args)
      throws InvalidConfigurationException,
             SolverException,
             InterruptedException {
    Configuration config = Configuration.fromCmdLineArguments(args);
    LogManager logger = BasicLogManager.create(config);
    ShutdownManager shutdown = ShutdownManager.create();

    System.out.println("Solving riddle ...");

    SolverContext context = SolverContextFactory.createSolverContext(
        config, logger, shutdown.getNotifier(), Solvers.PRINCESS);

    FormulaManager fmgr = context.getFormulaManager();

    BooleanFormulaManager bmgr = fmgr.getBooleanFormulaManager();
    IntegerFormulaManager imgr = fmgr.getIntegerFormulaManager();

    IntegerFormula h = imgr.makeVariable("h"),
                   a = imgr.makeVariable("a"),
                   t = imgr.makeVariable("t"),
                   e = imgr.makeVariable("e"),
                   l = imgr.makeVariable("l"),
                   o = imgr.makeVariable("o"),
                   v = imgr.makeVariable("v"),
                   s = imgr.makeVariable("s"),
                   k = imgr.makeVariable("k");

    List<IntegerFormula> allVars = new ArrayList<>();
    allVars.add(h);
    allVars.add(a);
    allVars.add(t);
    allVars.add(e);
    allVars.add(l);
    allVars.add(o);
    allVars.add(v);
    allVars.add(s);
    allVars.add(k);

    try (ProverEnvironment prover =
         context.newProverEnvironment(ProverOptions.GENERATE_MODELS)) {

      for (IntegerFormula x : allVars) {
          prover.addConstraint(imgr.greaterOrEquals(x, imgr.makeNumber(1)));
          prover.addConstraint(imgr.lessOrEquals(x, imgr.makeNumber(9)));
      }

      // better way to express the all-different constraint in Java-SMT?
      for (int i = 0; i < allVars.size(); ++i)
          for (int j = 0; j < i; ++j)
              prover.addConstraint(bmgr.not(imgr.equal(allVars.get(i),
                                                       allVars.get(j))));
      
      // h+a+t+e === 20
      // l+o+v+e === 14
      // s+o+l+v+e === 20
      // a+s+k === 14
      // l+e+a+s+t === 20
      // v+e+t === 14
      prover.addConstraint(imgr.equal(imgr.add(h, imgr.add(a, imgr.add(t, e))),
                                      imgr.makeNumber(20)));
      prover.addConstraint(imgr.equal(imgr.add(l, imgr.add(o, imgr.add(v, e))),
                                      imgr.makeNumber(14)));
      prover.addConstraint(imgr.equal(imgr.add(s, imgr.add(o,
                                             imgr.add(l, imgr.add(v, e)))),
                                      imgr.makeNumber(20)));
      prover.addConstraint(imgr.equal(imgr.add(a, imgr.add(s, k)),
                                      imgr.makeNumber(14)));
      prover.addConstraint(imgr.equal(imgr.add(l, imgr.add(e,
                                             imgr.add(a, imgr.add(s, t)))),
                                      imgr.makeNumber(20)));
      prover.addConstraint(imgr.equal(imgr.add(v, imgr.add(e, t)),
                                      imgr.makeNumber(14)));

      if (prover.isUnsat()) {
          System.out.println("No solution");
      } else {
          Model model = prover.getModel();
          for (IntegerFormula x : allVars)
              System.out.println("" + x + ": " + model.evaluate(x));

          BooleanFormula someDiff = bmgr.makeFalse();
          for (IntegerFormula x : allVars)
              someDiff =
                  bmgr.or(someDiff,
                          bmgr.not(imgr.equal(
                                      x,
                                      imgr.makeNumber(model.evaluate(x)))));

          prover.addConstraint(someDiff);

          if (prover.isUnsat())
              System.out.println("Solution is unique");
      }
    }
  }
}
